@@ -93,7 +93,16 @@ class Cursor(object):
         # TODO: make sure that conn.escape is correct
 
         if args is not None:
-            query = query % conn.escape(args)
+            if isinstance(args, tuple) or isinstance(args, list):
+                escaped_args = tuple(conn.escape(arg) for arg in args)
+            elif isinstance(args, dict):
+                escaped_args = dict((key, conn.escape(val)) for (key, val) in args.items())
+            else:
+                #If it's not a dictionary let's try escaping it anyways.
+                #Worst case it will throw a Value error
+                escaped_args = conn.escape(args)
+
+            query = query % escaped_args
 
         if isinstance(query, unicode):
             query = query.encode(charset)
@@ -113,12 +122,12 @@ class Cursor(object):
     def executemany(self, query, args):
         ''' Run several data against one query '''
         del self.messages[:]
-        conn = self._get_db()
+        #conn = self._get_db()
         if not args:
             return
-        charset = conn.charset
-        if isinstance(query, unicode):
-            query = query.encode(charset)
+        #charset = conn.charset
+        #if isinstance(query, unicode):
+        #    query = query.encode(charset)
 
         self.rowcount = sum([ self.execute(query, arg) for arg in args ])
         return self.rowcount
@@ -231,12 +240,9 @@ class Cursor(object):
         self.lastrowid = conn._result.insert_id
         self._rows = conn._result.rows
         self._has_next = conn._result.has_next
-        conn._result = None
 
     def __iter__(self):
-        self._check_executed()
-        result = self.rownumber and self._rows[self.rownumber:] or self._rows
-        return iter(result)
+        return iter(self.fetchone, None)
 
     Warning = Warning
     Error = Error
@@ -248,3 +254,44 @@ class Cursor(object):
     InternalError = InternalError
     ProgrammingError = ProgrammingError
     NotSupportedError = NotSupportedError
+
+class DictCursor(Cursor):
+    """A cursor which returns results as a dictionary"""
+
+    def execute(self, query, args=None):
+        result = super(DictCursor, self).execute(query, args)
+        if self.description:
+            self._fields = [ field[0] for field in self.description ]
+        return result
+
+    def fetchone(self):
+        ''' Fetch the next row '''
+        self._check_executed()
+        if self._rows is None or self.rownumber >= len(self._rows):
+            return None
+        result = dict(zip(self._fields, self._rows[self.rownumber]))
+        self.rownumber += 1
+        return result
+
+    def fetchmany(self, size=None):
+        ''' Fetch several rows '''
+        self._check_executed()
+        if self._rows is None:
+            return None
+        end = self.rownumber + (size or self.arraysize)
+        result = [ dict(zip(self._fields, r)) for r in self._rows[self.rownumber:end] ]
+        self.rownumber = min(end, len(self._rows))
+        return tuple(result)
+
+    def fetchall(self):
+        ''' Fetch all the rows '''
+        self._check_executed()
+        if self._rows is None:
+            return None
+        if self.rownumber:
+            result = [ dict(zip(self._fields, r)) for r in self._rows[self.rownumber:] ]
+        else:
+            result = [ dict(zip(self._fields, r)) for r in self._rows ]
+        self.rownumber = len(self._rows)
+        return tuple(result)
+
